'
'  PBcodec, Source Code Checker, v1.14
'-----------------------------------------------------------------------------
'  Scans a main source file and all its include files for Un-used Subs,
'  Functions, and variables. Version 1.1 also has ability to extract string
'  literals, that is, strings withing double-quotes.
'
'  PBcodec will save the result in a file ending with --LOG.TXT and also show
'  this file in associated texteditor.
'
'  This version does not process any Conditional Compiling statements, so far,
'  but all files are properly scanned and it even handles "DIM x AS GLOBAL in
'  a correct way. Tested on my own projects, all reports so far has been correct.
'
'  The output will show the SUB,FUNCT and VAR name followed by  [FileName: 565
'  where the last number is the line number where the item is declared. It will
'  also present a list of all Includes, Global vars, Subs and Functions, this
'  part originally written by Wayne Diamond.
'
'  Public Domain, this version by Borje Hagsten, July 2001, but main credits
'  goes to Scott Slater for showing us it could be done (and how to do it).
'  Parts of this program has been copied from his version, but parsing and
'  some other things has been totally rewritten, my way..
'  Many have been involved, giving many valuable tips and providing code,
'  so one could say this is the result of a joined PB community effort.. :)
'
'  Tip: Prog launches txt-associated texteditor to show report. Standard is
'  NotePad. I have set mine to use Courier New, 9p, which gives nice output..
'
'  A few notes: Exported subs/functions, or ones inside "public" include files,
'  may have a reference count of zero and still not be un-used, since some other
'  program may be using them.
'
'  Also, since one of the advantages with the PB language is almost unlimited
'  flexibility to do things, there's no guaranties everything will be found,
'  or even be 100% correct. It has proved to be successful on my projects,
'  but please, use this (and similar) programs with extreme care..
'
' LOG:
'  Aug 11, 2003 - added some extrs checks for line wraps, _ to DoProcess
'  Jan 23, 2002 - added support for relative include paths, plus optimized
'          some for better speed, etc.
'  Jan 23, 2002 - added support for relative include paths, plus optimized
'          some for better speed, etc.
'  Jan 17, 2002 - changed to use IsCharAlphaNumeric in ANSItrim, to include
'          leading/trailing digits in string literals. Also had to change
'          UCASER function a bit, so all now can be compiled in PBDLL 6.1 too..
'  Oct 17, Corrected error in DoSaveResults, where global/local name mix warning
'          could end up pointing at wrong file for first local declare.
'  Oct 10, added exclude() array to avoid some of the most common include file
'          names when extracting string literals. Possible to expand - see WinMain.
'          Also set string literal extraction checkbox as checked from start.
'  Oct 09, added possibility to extract string literals, that is, text within
'          double-quotes. Added AnsiTRIM function for this purpose.
'  Aug 01, in DoSaveResults, moved REDIM PRESERVE lVars out from loop in first
'          IF/THEN block and reversed loop, because it sometimes GPF'd there.
'          Of course it could GPF. Must run such loops backwards, silly me.
'  Aug 01, removed AllLocals array and use lVars to store all locals instead.
'          Changed report accordingly and now, Global/Local name mix lists
'          Line number for local representations too.
'  Jul 31, excluded declarations of Subs/Function from usage check
'  Jul 31, re-fixed previous stupid fix of GLOBAL DIM, so it works this time..
'  Jul 29, fixed code in DoProcess - ExtractLocals, to check DIM more carefully,
'          since DIM/REDIM may have been preceeded with a GLOBAL declare of same variable.
'  Jul 29, added support for multiple include file paths in WinMain and DoGetIncFiles
'          Added check for trailing backslash to fExist, so paths are handled correctly
'  Jul 29, added code to DoProcess - ExtractSub, to exclude declares for external
'          procedures (in DLL's etc.) from being counted as "declared but un-used".
'          Also added code to DoProcess - ExtractLine, to replace colons within
'          paranthesis, which could cause weird results when parsing a line for colons.
'  Jul 29, added "Scan" button to enable easy rescanning of a file, since I
'          have found this useful to do after changes have been made. Also did
'          some minor tweaking of the code to enhance performance.
'  Jul 28, major trimming of parser, to ensure results and improve performance.
'

#COMPILE EXE
'#INCLUDE "WIN32API.INC"  'Basic Win API definitions
#INCLUDE "PBCODEC.INC"  'Basic Win API definitions
'#INCLUDE "COMDLG32.INC"  'Common dialog's definitions

TYPE InfoStruct
   uName       AS ASCIIZ * 100 'is 100 enough? For me it is, but if not for you, increase..
   zName       AS ASCIIZ * 100
   inFunct     AS ASCIIZ * 100
   IsUsed      AS LONG
   iType       AS LONG
   FileNum     AS LONG
   LineNum     AS LONG
   SubEnd      AS LONG
   Exported    AS LONG
END TYPE

GLOBAL Vars()   AS InfoStruct  ' All Locals in Current Proc (TEMP)
GLOBAL gVars()  AS InfoStruct  ' All Globals
GLOBAL gDbl()   AS InfoStruct  ' Duplicate Global names
GLOBAL gDecl()  AS InfoStruct  ' Declared Subs/Functions
GLOBAL lVars()  AS InfoStruct  ' Un-Used Locals
GLOBAL Functs() AS InfoStruct  ' All Functions
GLOBAL Files()     AS STRING
GLOBAL inFile()    AS STRING
GLOBAL sIncDir()   AS STRING
GLOBAL sString()   AS STRING   'for string literals
GLOBAL exclude()   AS STRING   'for exclude strings
GLOBAL getStrings  AS LONG, sStrCount AS LONG

GLOBAL FilePathStr AS STRING, FileNameStr AS STRING, DestFile AS STRING
GLOBAL sTemp AS STRING, sWork AS STRING, LineStr AS STRING
GLOBAL I AS LONG, Done AS LONG

GLOBAL igDbl     AS LONG       ' # of Duplicate Globals
GLOBAL iVars     AS LONG       ' # of Vars
GLOBAL ilVars    AS LONG       ' # of lVars
GLOBAL igVars    AS LONG       ' # of gVars
GLOBAL iFuncts   AS LONG       ' # of Functs
GLOBAL DeclCount AS LONG
GLOBAL gTotLines AS LONG, t AS SINGLE

DECLARE CALLBACK FUNCTION WinMainProc() AS LONG
DECLARE FUNCTION AnsiTRIM(BYVAL txt AS STRING) AS STRING
DECLARE FUNCTION DoGetIncFiles(BYVAL TheFile AS STRING) AS LONG
DECLARE FUNCTION DoProcess(BYVAL TheFile AS STRING, BYVAL fNum AS LONG, WhatRun AS LONG) AS LONG
DECLARE FUNCTION GetCommandFile(BYVAL CmdStr AS STRING, Fi() AS STRING) AS LONG
DECLARE FUNCTION GetDroppedFile(BYVAL hDrop AS LONG, Fi() AS STRING) AS LONG
DECLARE FUNCTION GetIncludeDir AS STRING
DECLARE FUNCTION fExist(BYVAL zfName AS STRING) AS LONG
DECLARE FUNCTION IniGetString(BYVAL sSection AS STRING, BYVAL sKey AS STRING, _
                              BYVAL sDefault AS STRING, BYVAL sFile AS STRING) AS STRING
DECLARE FUNCTION IsFileMain(BYVAL fName AS STRING) AS LONG
DECLARE FUNCTION UCASER(BYVAL st AS STRING) AS STRING
DECLARE SUB DoInitProcess(BYVAL hDlg AS LONG, BYVAL fName AS STRING)
DECLARE SUB DoSaveResults
DECLARE SUB UCASES(a$)

'
' Main entrance - create dialog, etc.
'
FUNCTION WINMAIN (BYVAL hCurInstance  AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  lpszCmdLine         AS ASCIIZ PTR, _
                  BYVAL nCmdShow      AS LONG) AS LONG

  LOCAL hDlg AS LONG, iCnt AS LONG, rc AS RECT, tmpIncDir AS STRING, tmpStr AS STRING
  REDIM sIncDir(0)

  DIALOG NEW 0, "PBcodec v1.14", , , 165, 70, %WS_CAPTION OR %WS_MINIMIZEBOX OR %WS_SYSMENU TO hDlg
  IF hDlg = 0 THEN EXIT FUNCTION

  CONTROL ADD LABEL,   hDlg, 114, "",              2,  2, 161, 20, %SS_CENTER, %WS_EX_CLIENTEDGE
  CONTROL ADD LABEL,   hDlg, 115, " Main file: ",  2, 27, 161, 10
  CONTROL ADD BUTTON,  hDlg, 120, " &Browse..",    4, 53,  50, 14
  CONTROL ADD BUTTON,  hDlg, %IDOK, "&Scan",      58, 53,  50, 14
  CONTROL ADD BUTTON,  hDlg, %IDCANCEL, "&Quit", 112, 53,  50, 14
  CONTROL ADD CHECKBOX, hDlg, 130, "E&xtract string literals ", 4, 39, 80, 10
  CONTROL SET CHECK hDlg, 130, 1
  CONTROL DISABLE hDlg, %IDOK

  tmpIncDir = GetIncludeDir                              'grab include path from registry
  IF LEN(tmpIncDir) THEN                                 'if we got anything
     IF INSTR(tmpIncDir, ";") THEN                       'if it contains multiple paths
        FOR I = 1 TO PARSECOUNT(tmpIncDir, ";")          'loop through string
           tmpStr = TRIM$(PARSE$(tmpIncDir, ";", I))     'parse out each path
           IF LEN(tmpStr) AND fExist(tmpStr) THEN        'if we got a path and it exists
              REDIM PRESERVE sIncDir(iCnt)               'prepare array
              IF ASC(tmpStr, -1) = 92 THEN               'if a path with trailing backslash
                 sIncDir(iCnt) = tmpStr                  'store path in array element
              ELSE                                       'else
                  sIncDir(iCnt) = tmpStr + "\"           'make sure it has a trailing backslash
              END IF
              INCR iCnt                                  'increase temporary array counter
           END IF
        NEXT

     ELSE                                                'else, single path was given
        IF fExist(tmpIncDir) THEN                        'if it exists
           IF ASC(tmpIncDir, -1) = 92 THEN               'if a path with trailing backslash
              sIncDir(0) = tmpIncDir                     'store path in first array element
           ELSE                                          'else
              sIncDir(0) = tmpIncDir + "\"               'make sure it has a trailing backslash
           END IF
        END IF
     END IF
  END IF

  SystemParametersInfo %SPI_GETWORKAREA, BYVAL 0, BYVAL VARPTR(rc), 0      'grab desktop cordinates
  DIALOG PIXELS hDlg, rc.nRight, rc.nBottom TO UNITS rc.nRight, rc.nBottom 'convert to dialog units
  DIALOG SET LOC hDlg, rc.nRight - 185, rc.nBottom - 100                   'place dialog bottom, right
  SetWindowPos hDlg, %HWND_TOPMOST, 0, 0, 0, 0, %SWP_NOMOVE OR %SWP_NOSIZE 'set dialog topmost
  DragAcceptFiles     hDlg, %True                                          'enable drag&drop

  REDIM exclude(23) 'exclude these string literals
  exclude(0) = "WIN32API.INC"  : exclude(1) = "COMDLG32.INC"
  exclude(2) = "COMMCTRL.INC"  : exclude(3) = "DDT.INC"
  exclude(4) = "MDI32.INC"     : exclude(5) = "COMBO32.INC"
  exclude(6) = "LISTVIEW.INC"  : exclude(7) = "TRVIEW32.INC"
  exclude(8) = "RICHEDIT.INC"  : exclude(9) = "EDIT32.INC"
  exclude(10) = "BUTTON32.INC" : exclude(11) = "MMSYSTEM.INC"
  exclude(12) = "WSOCK32.INC"  : exclude(13) = "STATIC32.INC"
  exclude(14) = "DPMI.INC"     : exclude(15) = "LZEXPAND.INC"
  exclude(16) = "TOOLHLP.INC"  : exclude(17) = "VBAPI.INC"
  exclude(18) = "CTL3D.INC"    : exclude(19) = "VER.INC"
  exclude(20) = "WINAPI.INC"   : exclude(21) = "WINSOCK.INC"
  exclude(22) = "COMMDLG.INC"  : exclude(23) = "PROGRAM"

  DIALOG SHOW MODAL hDlg CALL WinMainProc
END FUNCTION

'
' Main callback procedure
'
CALLBACK FUNCTION WinMainProc() AS LONG

  SELECT CASE CBMSG
     CASE %WM_INITDIALOG
         REDIM Files(1)  ' Reset Array
         IF LEN(COMMAND$) THEN
             IF GetCommandFile(COMMAND$, Files()) THEN ' Retrieve the contents of the Command String
                LOCAL sTimer AS LONG
                sTimer = SETTIMER(CBHNDL, 1, 400, %NULL) ' wait for window to draw
             END IF
         ELSE
             CONTROL SET TEXT CBHNDL, 114, "Drag && Drop a main source file on dialog, " + $CRLF + _
                                           "or use Browse to select a file to Scan.."
         END IF

     CASE %WM_CTLCOLORSTATIC
        IF CBLPARAM = GetDlgItem(CBHNDL, 114) THEN
           SetBkColor CBWPARAM, GetSysColor(%COLOR_INFOBK)
           FUNCTION = GetSysColorBrush(%COLOR_INFOBK)
        END IF

     CASE %WM_TIMER
         KILLTIMER  CBHNDL, 1
         CALL DoInitProcess(CBHNDL, Files(0))

     CASE %WM_DROPFILES
         REDIM Files(1) ' Reset Array
         IF GetDroppedFile(CBWPARAM, Files()) THEN       ' Retrieve the Dropped filenames
            CALL DoInitProcess(CBHNDL, Files(0))
         END IF

     CASE %WM_DESTROY
        CALL DragAcceptFiles(CBHNDL, 0)

     CASE %WM_COMMAND
        SELECT CASE CBCTL
           CASE 120 'Browse
              LOCAL Style AS DWORD, fName AS STRING, Buffer AS STRING, Path AS STRING

              Style  = %OFN_EXPLORER OR %OFN_FILEMUSTEXIST OR %OFN_HIDEREADONLY
              fName  = "*.BAS"
              Buffer = "PB Code files (*.BAS)|*.BAS|"
              Path   = CURDIR$

              IF OpenFileDialog(CBHNDL, "", fName, Path, Buffer, "BAS" , Style) THEN
                 REDIM Files(1)  ' Reset Array
                 Files(0) = fName
                 fName = MID$(fName, INSTR(-1, fName, "\") + 1)
                 CONTROL SET TEXT CBHNDL, 115, " Main file: " & UCASER(fName)
                 CONTROL ENABLE CBHNDL, %IDOK
                 CALL DoInitProcess(CBHNDL, Files(0)) '<- deactivate, if not to scan directly..
              END IF

           CASE %IDOK
              REDIM PRESERVE Files(1)
              CALL DoInitProcess(CBHNDL, Files(0)) 'scan file

           CASE %IDCANCEL        ' Quit
               Done = 1          ' jump out of any loops
               DIALOG END CBHNDL ' and QUIT

        END SELECT
  END SELECT
END FUNCTION

'************************************************************************
' Initiate and run entire process
'************************************************************************
SUB DoInitProcess(BYVAL hDlg AS LONG, BYVAL fName AS STRING)
  LOCAL ci AS LONG, mc AS LONG
  FOR ci = 1 TO 10 : DIALOG DOEVENTS : NEXT

  IF fExist(fName) THEN 'make sure it exists
     ci = IsFileMain(fName)  'if a #COMPILE statement exists (main source file)

     IF ci = -3 THEN         'if return is -3, file was empty
        MSGBOX "Selected file is empty.",,"Error!"
        EXIT SUB

     ELSEIF ci = -2 THEN     'if return is -2, file could not be opened
        MSGBOX "Selected file could not be opened!",,"Error!"
        EXIT SUB

     ELSEIF ci >= 0 THEN     'if return is >= 0, file was not a main source file
        MSGBOX "Selected file does not contain a #COMPILE statement.",,"Error!"
        EXIT SUB
     END IF

  ELSE                       'else, it didn't even exist..
     MSGBOX "Could not open this file:" + $CRLF + fName + $CRLF + $CRLF + _
            "Please make sure it exists and try again.",,"Error!"
     EXIT SUB
  END IF

  CONTROL GET CHECK hDlg, 130 TO getStrings
  CONTROL DISABLE hDlg, 120
  CONTROL DISABLE hDlg, %IDOK
  CONTROL DISABLE hDlg, 130

  FilePathStr = LEFT$(fName, INSTR(-1, fName, "\"))
  FileNameStr = MID$(fName, INSTR(-1, fName, "\") + 1)
  CONTROL SET TEXT hDlg, 115, " Main file: " & UCASER(FileNameStr)

  REDIM inFile(0)
  IF Files(0) = "" THEN Files(0) = fName
  CHDRIVE LEFT$(FilePathStr, 2)
  CHDIR FilePathStr

  CONTROL SET TEXT hDlg, 114, "Collecting include files"
  DoGetIncFiles fName

  iFuncts = 0 : REDIM Functs(0)
  igVars  = 0 : REDIM gVars(0)
  ilVars  = 0 : REDIM lVars(0)
  iVars   = 0 : REDIM Vars(0)
  igDbl   = 0 : REDIM gDbl(0)
  DeclCount = 0 : REDIM gDecl(0)
  sStrCount = 0 : REDIM sString(0)
  gTotLines = 0
  t = TIMER

  FOR mc = 0 TO 1
     FOR ci = 0 TO UBOUND(Files) - 1
        SELECT CASE UCASE$(MID$(Files(ci), INSTR(-1, Files(ci), "\") + 1))       'ignore these
           CASE "WIN32API.INC", "COMDLG32.INC", "COMMCTRL.INC", "COMBO32.INC", _
                "DDT.INC", "MDI32.INC", "LISTVIEW.INC", "TRVIEW32.INC", "RICHEDIT.INC", _
                "EDIT32.INC", "BUTTON32.INC", "MMSYSTEM.INC", "WSOCK32.INC", _
                "STATIC32.INC", "DPMI.INC", "LZEXPAND.INC", "TOOLHLP.INC", "VBAPI.INC", _
                "CTL3D.INC", "VER.INC", "WINAPI.INC", "WINSOCK.INC", "COMMDLG.INC"
           CASE ELSE
              IF mc = 0 THEN
                 CONTROL SET TEXT hDlg, 114, "Scanning for Local vars, Subs and Functions in: " + _
                             MID$(Files(ci), INSTR(-1, Files(ci), "\") + 1)
              ELSE
                 CONTROL SET TEXT hDlg, 114, "Scanning for Global vars in: " + _
                             MID$(Files(ci), INSTR(-1, Files(ci), "\") + 1)
              END IF
              DoProcess Files(ci), ci, mc
        END SELECT
     NEXT
     IF mc = 0 THEN
        IF iFuncts THEN REDIM PRESERVE Functs(iFuncts)
        IF igVars  THEN REDIM PRESERVE gVars(igVars)
        IF ilVars  THEN REDIM PRESERVE lVars(ilVars)
        IF igDbl   THEN REDIM PRESERVE gDbl(igDbl)
        IF DeclCount THEN REDIM PRESERVE gDecl(DeclCount)
        IF sStrCount THEN REDIM PRESERVE sString(sStrCount)
     END IF
  NEXT
  t = TIMER - t

  CONTROL ENABLE hDlg, 120
  CONTROL ENABLE hDlg, %IDOK
  CONTROL ENABLE hDlg, 130
  CONTROL SET TEXT hDlg, 114, "Done! Drag && Drop a main source file on dialog, " + $CRLF + _
                              "or use Browse to select a file to Scan.."
  CALL DoSaveResults

END SUB

'
' Get all included files into array
'
FUNCTION DoGetIncFiles(BYVAL TheFile AS STRING) AS LONG
  LOCAL ci AS LONG, ii AS LONG, p AS LONG, fc AS LONG, Letter AS BYTE PTR, sTemp2 AS STRING
  REDIM tmpFiles(0) AS STRING

  OPEN TheFile FOR BINARY AS #1 LEN = 16383
     IF ERR THEN
        RESET : ERRCLEAR : EXIT FUNCTION
     END IF
     GET$ #1, LOF(1), sTemp
  CLOSE #1
  IF LEN(sTemp) < 2 THEN EXIT FUNCTION

  p = 1: ci = 0
  Letter = STRPTR(sTemp)

  FOR ci = 1 TO LEN(sTemp)
     IF @Letter = 13 THEN
        IF ci - p > 2 THEN                             'if more than a line feed
           LineStr = UCASE$(LTRIM$(MID$(sTemp, p, ci - p) ))
           IF (ASC(LineStr) = 36 OR ASC(LineStr) = 35) AND _  '$#
           MID$(LineStr, 2, 8) = "INCLUDE " THEN       'can be several INCLUDE statements in file, anywhere in code..
              sWork = PARSE$(LineStr, CHR$(34), 2)     'get filename

              IF LEFT$(sWork, 2) = ".\" THEN  'resolve eventual relative paths
                 sWork = FilePathStr + MID$(sWork, 2)
              ELSEIF LEFT$(sWork, 3) = "..\" THEN
                 sWork = LEFT$(FilePathStr, INSTR(-2, FilePathStr, "\")) + MID$(sWork, 4)
              ELSEIF LEFT$(sWork, 4) = "...\" THEN
                 sTemp2 = LEFT$(FilePathStr, INSTR(-2, FilePathStr, "\"))
                 sWork = LEFT$(sTemp2, INSTR(-2, sTemp2, "\")) + MID$(sWork, 5)
              ELSEIF LEFT$(sWork, 5) = "....\" THEN
                 sTemp2 = LEFT$(FilePathStr, INSTR(-2, FilePathStr, "\"))
                 sTemp2 = LEFT$(sTemp2, INSTR(-2, sTemp2, "\"))
                 sWork = LEFT$(sTemp2, INSTR(-2, sTemp2, "\")) + MID$(sWork, 6)
              ELSEIF LEFT$(sWork, 6) = ".....\" THEN
                 sTemp2 = LEFT$(FilePathStr, INSTR(-2, FilePathStr, "\"))
                 sTemp2 = LEFT$(sTemp2, INSTR(-2, sTemp2, "\"))
                 sTemp2 = LEFT$(sTemp2, INSTR(-2, sTemp2, "\"))
                 sWork = LEFT$(sTemp2, INSTR(-2, sTemp2, "\")) + MID$(sWork, 7)
              ELSEIF LEFT$(sWork, 7) = "......\" THEN
                 sTemp2 = LEFT$(FilePathStr, INSTR(-2, FilePathStr, "\"))
                 sTemp2 = LEFT$(sTemp2, INSTR(-2, sTemp2, "\"))
                 sTemp2 = LEFT$(sTemp2, INSTR(-2, sTemp2, "\"))
                 sTemp2 = LEFT$(sTemp2, INSTR(-2, sTemp2, "\"))
                 sWork = LEFT$(sTemp2, INSTR(-2, sTemp2, "\")) + MID$(sWork, 8)
              END IF

              IF INSTR(-1, sWork, ".") = 0 THEN          'if no file extension is given,
                 sWork = sWork + ".BAS"                  'compiler assumes .BAS file
              END IF

              'if no path is given, compiler will first look in include dir, so we better start there
              IF INSTR(sWork, "\") = 0 THEN                          'if no path given
                 FOR ii = 0 TO UBOUND(sIncDir)                       'loop through the ones we have
                    IF fExist(sIncDir(ii) + sWork) THEN              'if file exists
                       sWork = sIncDir(ii) + sWork                   'use it
                    END IF
                 NEXT
              END IF

              IF LEN(FilePathStr) AND fExist(FilePathStr + sWork) THEN 'try with current file's path
                 sWork = FilePathStr + sWork                      'if ok, use it
              END IF

              IF fExist(sWork) THEN                     'safety check - if we can find what we got..
                 tmpFiles(fc) = sWork                   'store path + name in temporary array
                 INCR fc : REDIM PRESERVE tmpFiles(fc)  'incr counter and redim temporary array
              END IF
           END IF
        END IF
        INCR Letter : INCR ci : p = ci + 1              'jump past LF and store position
     END IF
     INCR Letter
  NEXT ci

  IF fc THEN
     p = UBOUND(Files)
     REDIM PRESERVE Files(p + fc)
     REDIM PRESERVE inFile(p + fc)

     FOR ci = 0 TO fc - 1
        Files(p + ci)  = tmpFiles(ci)
        inFile(p + ci) = MID$(TheFile, INSTR(-1, TheFile, "\") + 1)
     NEXT
     FOR ci = 0 TO fc - 1
        SELECT CASE UCASE$(MID$(tmpFiles(ci), INSTR(-1, tmpFiles(ci), "\") + 1))   'ignore these
           CASE "WIN32API.INC", "COMDLG32.INC", "COMMCTRL.INC", "COMBO32.INC", _
                "DDT.INC", "MDI32.INC", "LISTVIEW.INC", "TRVIEW32.INC", "RICHEDIT.INC", _
                "EDIT32.INC", "BUTTON32.INC", "MMSYSTEM.INC", "WSOCK32.INC", _
                "STATIC32.INC", "DPMI.INC", "LZEXPAND.INC", "TOOLHLP.INC", "VBAPI.INC", _
                "CTL3D.INC", "VER.INC", "WINAPI.INC", "WINSOCK.INC", "COMMDLG.INC"

           CASE ELSE
              CALL DoGetIncFiles(tmpFiles(ci)) 'recursive call to get eventual includes in includes
         END SELECT
     NEXT
  END IF

END FUNCTION

'
' Trim away all leading/ending non-letters and digits from a string
'
FUNCTION AnsiTRIM(BYVAL txt AS STRING) AS STRING
  LOCAL pos1 AS LONG, pos2 AS LONG

  FOR pos1 = 1 TO LEN(txt)
     IF IsCharAlphaNumeric(ASC(txt, pos1)) THEN EXIT FOR
  NEXT
  FOR pos2 = LEN(txt) TO 1 STEP -1
     IF IsCharAlphaNumeric(ASC(txt, pos2)) THEN EXIT FOR
  NEXT

  IF Pos2 > Pos1 THEN FUNCTION = MID$(txt, Pos1, Pos2 - Pos1 + 1)

END FUNCTION

'
' Load text from file, extract lines and get all subs, functions and globals
' into arrays.
'
FUNCTION DoProcess(BYVAL TheFile AS STRING, BYVAL fNum AS LONG, WhatRun AS LONG) AS LONG
  IF fExist(TheFile) = 0 THEN EXIT FUNCTION  'if file doesn't exist, exit

  LOCAL ci AS LONG, p AS LONG, Letter AS BYTE PTR, Letter2 AS BYTE PTR
  LOCAL Ac AS LONG, K AS LONG, sFlag AS LONG, QuotePos AS LONG, QuotePos2 AS LONG, dbl AS LONG, dUscored AS LONG
  LOCAL exported AS LONG, x AS LONG, y AS LONG, uscoredGlobal AS LONG, endRout AS LONG, fUscored AS LONG
  LOCAL di AS LONG, pos AS LONG, wordFlag AS LONG, StrFlag AS LONG, dFlag AS LONG, inPar AS LONG
  LOCAL locX AS LONG, locY AS LONG, locPos AS LONG, uscoredLocal AS LONG, iv AS LONG, isGLOBAL AS LONG
  LOCAL MainStr AS STRING, sBuf AS STRING, txt AS STRING, Buf AS STRING, fsName AS STRING, strDump AS STRING

'--------------------------------------------------------------------
' load file into MainStr
'--------------------------------------------------------------------
  OPEN TheFile FOR BINARY AS #1 LEN = 16384    'open file
     IF ERR THEN                               'if it failed
        RESET : ERRCLEAR : EXIT FUNCTION       'reset, clear error and exit
     END IF
     GET$ #1, LOF(1), MainStr                  'else load contents into MainStr
  CLOSE #1                                     'close file
  IF LEN(MainStr) < 2 THEN EXIT FUNCTION       'if MainStr is too short, exit

'--------------------------------------------------------------------
' scan MainStr and extract lines
'--------------------------------------------------------------------
  p = 1: ci = 0 : I = 0
  Letter = STRPTR(MainStr)                       'point to beginning of MainStr

  FOR ci = 1 TO LEN(MainStr)
     SELECT CASE @Letter
        CASE 9  : @Letter = 32                   'replace ev. tab with space
        CASE 13
           INCR I : INCR gTotLines               'increase line- and total line counter
           IF ci - p > 3 THEN GOSUB ExtractLine  'if line is enough long, extract it
           INCR Letter : INCR ci : p = ci + 1    'jump past LF and store position
     END SELECT
     INCR Letter
  NEXT ci

  IF ci > p + 2 THEN
     INCR I                    'increase line counter
     GOSUB ExtractLine         'if line didn't end with line feed..
  END IF
EXIT FUNCTION

'---------------------------------------------------------
' Extract line from main text
'---------------------------------------------------------
ExtractLine:
  txt = MID$(MainStr, p, ci - p)               'extract line
  IF ASC(LTRIM$(txt)) = 39 THEN RETURN         'if whole line is uncommented, get next

  '--------------------------------------------------------------------
  ' blank out text within double quotes
  '--------------------------------------------------------------------
  QuotePos = INSTR(txt, $DQ)                   'see if there is any
  IF QuotePos THEN
     DO                                        'loop while there is any left
        QuotePos2 = INSTR(QuotePos + 1, txt, $DQ) 'look for matching pair
        IF QuotePos2 THEN
           IF WhatRun = 0 AND getStrings = 1 THEN 'if to extract string literals
              strDump = AnsiTRIM(MID$(txt, QuotePos, QuotePos2 - QuotePos + 1))
              IF LEN(strDump) THEN
                 ARRAY SCAN exclude(), FROM 1 TO LEN(strDump), COLLATE UCASE, = strDump, TO Ac
                 IF Ac = 0 THEN
                    IF sStrCount MOD 20 = 0 THEN REDIM PRESERVE sString(sStrCount + 20)
                    sString(sStrCount) = FORMAT$(fNum) + $TAB + USING$("####", I) + $TAB  + strDump
                    INCR sStrCount
                 END IF
              END IF
           END IF
           MID$(txt, QuotePos, QuotePos2 - QuotePos + 1) = SPACE$(QuotePos2 - QuotePos + 1)
           QuotePos = INSTR(QuotePos2 + 1, txt, $DQ)
           IF QuotePos = 0 THEN EXIT DO
        ELSE
           EXIT DO
        END IF
     LOOP
  END IF

  txt = LTRIM$(EXTRACT$(txt, "'"))              'cut off ev. uncommented part and trim away leading spaces
  IF ASC(txt) = 82 AND LEFT$(txt, 4) = "REM " THEN RETURN 'if whole line in uncommented
  IF INSTR(UCASE$(txt), " REM ") THEN
     txt = LEFT$(txt, INSTR(UCASE$(txt), "REM ")) 'same here, if REM was used
  END IF
  IF INSTR(txt, " _") THEN
     txt = LEFT$(txt, INSTR(txt, " _") + 2)      'if line wraps to next, ignore the rest of it
  ELSEIF INSTR(txt, ",_ ") THEN
     txt = LEFT$(txt, INSTR(txt, ",_ ")) + " _"  'adjust to parser
  ELSEIF RIGHT$(txt, 2) = ",_" THEN
     txt = LEFT$(txt, LEN(txt) - 2) + " _"       'adjust to parser
  END IF
  IF ASC(txt, -1) = 32 THEN txt = RTRIM$(txt)       'trim off trailing spaces

  IF INSTR(txt, ":") THEN                       'colon inside paranthesis must be converted
     inPar = 0
     FOR Letter2 = STRPTR(txt) TO STRPTR(txt) + LEN(txt)
        SELECT CASE @Letter2
           CASE 40 : INCR inPar                 'left paranthesis (
           CASE 41 : DECR inPar                 'right paranthesis )
           CASE 58 : IF inPar > 0 THEN @Letter2 = 59 'if within paranthesis, convert colon to semicolon (whatever)
        END SELECT
     NEXT
  END IF

  IF LEN(txt) > 2 THEN                          'now, if line is enough long
     IF WhatRun = 0 THEN                        'and first run
        GOSUB ExtractSub                        'send it to sub/function check
        GOSUB ExtractGlobal                     'send it to global variable check
     ELSE
        IF LTRIM$(UCASE$(LEFT$(txt, 8))) = "DECLARE " OR dUscored THEN
           dUscored = (ASC(TRIM$(txt), -1) = 95)
           RETURN
        END IF
        GOSUB ChkVariables                      'second run, calculate globals
     END IF
  END IF

RETURN

'---------------------------------------------------------
' Get subs and functions, plus get/check local variables (DIM, LOCAL, STATIC)
'---------------------------------------------------------
ExtractSub:
  IF sFlag = 0 THEN
     Buf = UCASER(txt)

     IF LEFT$(Buf, 8) = "DECLARE " THEN       'Declaration
        IF INSTR(Buf, " LIB ") THEN RETURN                'external routine - DLL, etc.
        txt = LTRIM$(MID$(txt, 9))
        Buf = UCASE$(txt) : dFlag = 1
        IF LEN(RTRIM$(txt)) = 1 AND ASC(txt) = 95 THEN RETURN
     END IF

     IF LEFT$(Buf, 9) = "FUNCTION " THEN                   'Function start
        sFlag = 2 : fsName = LTRIM$(MID$(txt, 10)) : K = I
     ELSEIF LEFT$(Buf, 4) = "SUB " THEN                    'Sub start
        sFlag = 1 : fsName = LTRIM$(MID$(txt, 5))  : K = I
     ELSEIF LEFT$(Buf, 18) = "CALLBACK FUNCTION " THEN     'Callback Function start
        sFlag = 3 : fsName = LTRIM$(MID$(txt, 19)) : K = I
     END IF

     IF sFlag THEN
        IF INSTR(UCASE$(fsName), " EXPORT") THEN exported = 1
        Ac = INSTR(fsName, ANY " ('")
        IF Ac THEN fsName = TRIM$(LEFT$(fsName, Ac - 1), ANY " &%@!#$?")
     END IF
     IF LEN(fsName) = 1 AND ASC(fsName) = 95 THEN fUscored = 1

     IF dFlag AND fUscored = 0 THEN
        GOSUB AddDeclare : RETURN
     END IF

  ELSE
     IF fUscored THEN
        IF dFlag AND INSTR(Buf, " LIB ") THEN RETURN             'external routine, DLL
        IF fUscored = 1 THEN 'look for name
           ac = INSTR(LTRIM$(txt), ANY " (")
           IF Ac THEN
              fsName = TRIM$(LEFT$(txt, Ac - 1), ANY " &%@!#$?")
           ELSE
              fsName = TRIM$(txt, ANY " &%@!#$?")
           END IF
        ELSE
           IF INSTR(UCASE$(txt), "EXPORT") THEN exported = 1
        END IF
        IF ASC(TRIM$(txt), -1) = 95 THEN
           fUscored = 2
        ELSE
           fUscored = 0
        END IF
     END IF

     IF dFlag AND fUscored = 0 THEN 'declaration
        GOSUB AddDeclare : RETURN
     END IF

     SELECT CASE sFlag
        CASE 1
           IF UCASE$(LEFT$(txt, 7)) = "END SUB" THEN
              endRout = sFlag
           ELSE
              GOSUB ExtractLocals
              GOSUB ChkVariables
              RETURN
           END IF
        CASE 2, 3
           IF UCASE$(LEFT$(txt, 12)) = "END FUNCTION" THEN
              endRout = sFlag
           ELSE
              GOSUB ExtractLocals
              GOSUB ChkVariables
              RETURN
           END IF
     END SELECT

     IF endRout THEN
        IF iFuncts MOD 40 = 0 THEN REDIM PRESERVE Functs(iFuncts + 40)
        fsName = RTRIM$(fsName)
        Functs(iFuncts).zName    = fsName
        CALL UCASES(fsName)
        Functs(iFuncts).uName    = fsName & CHR$(0)
        Functs(iFuncts).iType    = endRout
        Functs(iFuncts).LineNum  = K
        Functs(iFuncts).SubEnd   = I
        Functs(iFuncts).FileNum  = fNum
        Functs(iFuncts).Exported = exported
        INCR iFuncts
        sFlag = 0 : endRout = 0 : exported = 0

        IF iVars THEN
           REDIM PRESERVE lVars(ilVars + iVars)
           FOR iv = 0 TO iVars - 1
              lVars(ilVars).zName   = Vars(iv).zName
              lVars(ilVars).uName   = Vars(iv).uName
              lVars(ilVars).InFunct = Vars(iv).InFunct
              lVars(ilVars).iType   = Vars(iv).iType
              lVars(ilVars).LineNum = Vars(iv).LineNum
              lVars(ilVars).FileNum = Vars(iv).FileNum
              lVars(ilVars).IsUsed  = Vars(iv).IsUsed
              INCR ilVars
           NEXT
           iVars = 0 : REDIM Vars(0)
        END IF
     END IF
  END IF

RETURN

AddDeclare:
  IF DeclCount MOD 40 = 0 THEN REDIM PRESERVE gDecl(DeclCount + 40)
  fsName = RTRIM$(fsName)
  gDecl(DeclCount).zName    = fsName
  CALL UCASES(fsName)
  gDecl(DeclCount).uName    = fsName & CHR$(0)
  gDecl(DeclCount).iType    = sFlag
  gDecl(DeclCount).LineNum  = K
  gDecl(DeclCount).SubEnd   = I
  gDecl(DeclCount).FileNum  = fNum
  gDecl(DeclCount).Exported = exported
  INCR DeclCount
  sFlag = 0 : endRout = 0 : exported = 0 : dFlag = 0
RETURN

'---------------------------------------------------------
' Get Locals
'---------------------------------------------------------
ExtractLocals:
  IF INSTR(UCASE$(txt), "LOCAL ") OR INSTR(UCASE$(txt), "DIM ") OR _
     INSTR(UCASE$(txt), "STATIC ") OR uscoredLocal THEN
     FOR locX = 1 TO PARSECOUNT(txt, ":")
        sWork = TRIM$(PARSE$(txt, ":", locX))
        IF UCASE$(LEFT$(sWork, 6)) = "LOCAL " OR _
           UCASE$(LEFT$(sWork, 4)) = "DIM " OR _
           UCASE$(LEFT$(sWork, 7)) = "STATIC " OR _
           uscoredLocal THEN

            IF uscoredLocal = 0 THEN
               IF UCASE$(LEFT$(sWork, 6)) = "LOCAL " THEN
                  isGLOBAL = 0 : sWork = MID$(sWork, 7)
               ELSEIF UCASE$(LEFT$(sWork, 4)) = "DIM " THEN
                  isGLOBAL = 1 : sWork = MID$(sWork, 5)    'start out by assuming global status
               ELSEIF UCASE$(LEFT$(sWork, 7)) = "STATIC " THEN
                  isGLOBAL = 0 : sWork = MID$(sWork, 8)
               END IF
            END IF
            FOR locY = 1 TO PARSECOUNT(sWork, ",")
                sBuf = TRIM$(PARSE$(sWork, ",", locY))

                IF isGLOBAL = 1 THEN 'check if DIM statement really was global
                   IF INSTR(UCASE$(sBuf), " GLOBAL") THEN    'this can only happen
                      isGLOBAL = 2                           'with "DIM xx AS GLOBAL.."
                   ELSEIF INSTR(UCASE$(sBuf), " LOCAL") OR _ 'local DIM..
                         INSTR(UCASE$(sBuf), " STATIC") THEN
                      isGLOBAL = 0
                   END IF
                END IF

                sBuf = EXTRACT$(sBuf, ANY " ()")  'Chop off AS LONG, etc, or if array - (
                sBuf = RTRIM$(sBuf, ANY " &%@!#$?")
                IF LEN(sBuf) = 1 AND ASC(sBuf) = 95 THEN ITERATE
                sBuf = sBuf + CHR$(0)

                IF isGLOBAL < 2 THEN
                   ARRAY SCAN Vars(), FROM 1 TO LEN(sBuf), COLLATE UCASE, = sBuf, TO locPos
                   IF locPos = 0 THEN
                      IF iVars MOD 40 = 0 THEN REDIM PRESERVE Vars(iVars + 40)
                      Vars(iVars).zName   = sBuf
                      Vars(iVars).uName   = UCASER(sBuf)
                      Vars(iVars).InFunct = fsName
                      Vars(iVars).FileNum = fNum
                      Vars(iVars).iType   = isGlobal
                      Vars(iVars).LineNum = I
                      INCR iVars
                   END IF

                ELSE
                   ARRAY SCAN gVars(), FROM 1 TO LEN(sBuf), COLLATE UCASE, = sBuf, TO pos
                   IF pos = 0 THEN 'if not already there, add it (GLOBAL+DIM/REDIM, DIM AS GLOBAL, etc.)
                      IF igVars MOD 40 = 0 THEN REDIM PRESERVE gVars(igVars + 40)
                      gVars(igVars).zName = sBuf
                      gVars(igVars).uName = UCASER(sBuf)
                      gVars(igVars).FileNum = fNum
                      gVars(igVars).LineNum = I
                      INCR igVars
                   END IF

                END IF
            NEXT

        END IF
     NEXT
     uscoredLocal = (RIGHT$(RTRIM$(txt), 2)  = " _")
  END IF

RETURN

'---------------------------------------------------------
' Get Globals
'---------------------------------------------------------
ExtractGlobal:
  IF INSTR(UCASE$(txt), "GLOBAL ") OR uscoredGlobal THEN
     FOR x = 1 TO PARSECOUNT(txt, ":")
        sWork = TRIM$(PARSE$(txt, ":", x))
        isGLOBAL = (UCASE$(LEFT$(sWork, 7)) = "GLOBAL ")
        IF isGLOBAL = 0 THEN isGLOBAL = uscoredGlobal
        IF UCASE$(LEFT$(sWork, 7)) = "GLOBAL " OR uscoredGlobal THEN
            IF uscoredGlobal = 0 THEN sWork = MID$(sWork, 8)
            FOR y = 1 TO PARSECOUNT(sWork, ",")
                sBuf = TRIM$(PARSE$(sWork, ",", y))
                sBuf = EXTRACT$(sBuf, ANY " ()")  'Chop off AS LONG etc.
                sBuf = RTRIM$(sBuf, ANY " &%@!#$?")
                IF LEN(sBuf) = 1 AND ASC(sBuf) = 95 THEN ITERATE
                sBuf = sBuf + CHR$(0)

                IF igVars THEN 'must check for ev. duplicate declarations
                   ARRAY SCAN gVars(), FROM 1 TO LEN(sBuf), COLLATE UCASE, = sBuf, TO dbl
                   IF dbl THEN
                      IF igDbl THEN
                         ARRAY SCAN gDbl(), FROM 1 TO LEN(sBuf), COLLATE UCASE, = sBuf, TO dbl
                         IF dbl THEN
                            INCR gDbl(dbl - 1).IsUsed
                            ITERATE FOR
                         END IF
                      END IF
                      REDIM PRESERVE gDbl(igDbl)
                      gDbl(igDbl).zName   = sBuf
                      gDbl(igDbl).uName   = UCASER(sBuf)
                      gDbl(igDbl).FileNum = fNum
                      gDbl(igDbl).LineNum = I
                      INCR igDbl
                      ITERATE FOR
                   END IF
                END IF

                IF igVars MOD 40 = 0 THEN REDIM PRESERVE gVars(igVars + 40)
                gVars(igVars).zName = sBuf
                gVars(igVars).uName = UCASER(sBuf)
                gVars(igVars).FileNum = fNum
                gVars(igVars).LineNum = I
                INCR igVars
            NEXT

        END IF
     NEXT
     IF isGlobal THEN uscoredGlobal = ( RIGHT$(RTRIM$(txt), 2)  = " _" )
  END IF

RETURN

'---------------------------------------------------------
' Check variables
'---------------------------------------------------------
ChkVariables:
  wordFlag = 0 : StrFlag = 0
  Letter2 = STRPTR(txt)
  FOR di = 1 TO LEN(txt)
     SELECT CASE @Letter2
        CASE 97 TO 122, 65 TO 90, 48 TO 57, 95, 192 TO 214, 216 TO 246, 248 TO 255
           IF wordFlag = 0 AND @Letter2 <> 95 THEN 'if valid char and no flag, word starts here (not with underscore)
              wordFlag = 1 : pos = di              'set wordflag and store position
           END IF

        CASE ELSE                       'we hit something else, like space, dot, etc..
           IF wordFlag = 1 THEN         'if flag, then a word is ready
              GOSUB ChkWord             'check what we got
              wordFlag = 0              'and reset wordflag
           END IF
     END SELECT
     INCR Letter2                       'next char
  NEXT

  IF wordFlag  = 1 THEN GOSUB ChkWord 'in case there were letters all the way to the end..
RETURN

ChkWord:
   pos = di - pos                             'calculate length
   sBuf = PEEK$(Letter2 - pos, pos) + CHR$(0) 'grab word
   CALL UCASES(sBuf)                          'make it upper case for compare

  IF WhatRun = 0 THEN                         'check local variables
     ARRAY SCAN Vars(), FROM 1 TO LEN(sBuf), = sBuf, TO pos
     IF pos THEN INCR Vars(pos - 1).IsUsed

  ELSE                                        'check Subs(Functions and Global vars
     ARRAY SCAN Functs(), FROM 1 TO LEN(sBuf), = sBuf, TO pos
     IF pos THEN INCR Functs(pos - 1).IsUsed

     ARRAY SCAN gVars(), FROM 1 TO LEN(sBuf), = sBuf, TO pos
     IF pos THEN INCR gVars(pos - 1).IsUsed
  END IF

RETURN

END FUNCTION

'
' Prepare and save a report of what we've found out!
'
SUB DoSaveResults
  LOCAL hFile AS LONG, lRes AS LONG, fDbl AS LONG, cTmp AS LONG, _
        uline AS STRING, fName AS STRING, aStr AS STRING, aStr2 AS STRING

'------------------------------------------------------------------------------
' first look through locals array. if iType = 1, it may be a GLOBAL DIM
'------------------------------------------------------------------------------
  IF igVars AND ilVars THEN
     FOR I = ilVars - 1 TO 0 STEP -1 '<- must run this backwards through array!
        IF lVars(I).iType = 1 THEN
           aStr = lVars(I).uName + CHR$(0)
           ARRAY SCAN gVars(), FROM 1 TO LEN(aStr), = aStr, TO cTmp
           IF cTmp THEN               'if also GLOBAL, remove from local arrays
              ARRAY DELETE lVars(I)
              DECR ilVars
           END IF
        END IF
     NEXT
  END IF
  REDIM PRESERVE lVars(ilVars)

'------------------------------------------------------------------------------
' now prepare report..
'------------------------------------------------------------------------------
  DestFile = PARSE$(FileNameStr, ANY ".", 1)+"LOG.txt"
  OPEN FilePathStr + DestFile FOR OUTPUT AS hFile
  sWork = STRING$(80,"")
  uline = STRING$(80,"-")

'------------------------------------------------------------------------------
  PRINT# hFile, sWork
  PRINT# hFile, " PBcodec report: "  UCASER(FileNameStr) + " + include files. " & _
                "Generated " & DATE$ & ", " & TIME$
  PRINT# hFile, STR$(gTotLines) + " lines scanned in " + FORMAT$(t, "0.000") + _
                " seconds (" + FORMAT$(gTotLines / t * 60, "0") + " lines/minute)"
  PRINT# hFile, sWork
'------------------------------------------------------------------------------
  IF UBOUND(Files) > -1 THEN
     PRINT# hFile, " MAIN + INCLUDE FILES"
     PRINT# hFile, uline
     inFile(0) = "Main source file"
     FOR I = 0 TO UBOUND(Files) - 1
        SELECT CASE UCASE$(MID$(Files(I), INSTR(-1, Files(I), "\") + 1))   'these have been ignored
           CASE "WIN32API.INC", "COMDLG32.INC", "COMMCTRL.INC", "COMBO32.INC", _
                "DDT.INC", "MDI32.INC", "LISTVIEW.INC", "TRVIEW32.INC", "RICHEDIT.INC", _
                "EDIT32.INC", "BUTTON32.INC", "MMSYSTEM.INC", "WSOCK32.INC", _
                "STATIC32.INC", "DPMI.INC", "LZEXPAND.INC", "TOOLHLP.INC", "VBAPI.INC", _
                "CTL3D.INC", "VER.INC", "WINAPI.INC", "WINSOCK.INC", "COMMDLG.INC"

              aStr = " (not scanned)"
           CASE ELSE
              aStr = ""
        END SELECT

        PRINT# hFile, " " & LEFT$(Files(I) & aStr & SPACE$(58), 58) & "[" +inFile(I) + "]"
     NEXT I
  END IF
'------------------------------------------------------------------------------
  PRINT# hFile, ""
  PRINT# hFile, sWork
  PRINT# hFile, " UN-USED FUNCTIONS  (exported, or in incl. files, may be used by other programs)"
  PRINT# hFile, uline
  IF iFuncts THEN
     FOR I = 0 TO iFuncts - 1
         IF Functs(I).IsUsed = 1 AND Functs(I).iType > 1 THEN
            SELECT CASE UCASE$(Functs(I).zName)
               CASE "PBMAIN", "WINMAIN", "LIBMAIN", "PBLIBMAIN", "DLLMAIN" 'ignore these
               CASE ELSE
                  fName = Files(Functs(I).FileNum)
                  aStr = " FUNCTION " : aStr2 = ""
                  IF Functs(I).iType = 3 THEN aStr = " CALLBACK "
                  IF Functs(I).Exported THEN aStr2 = " <EXPORT>"
                  PRINT# hFile, LEFT$(aStr & Functs(i).zName & aStr2 & SPACE$(52), 52) & "  [" & _
                         MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" STR$(Functs(I).LineNum)
            END SELECT
        END IF
     NEXT I
  END IF
'------------------------------------------------------------------------------
  PRINT# hFile, ""
  PRINT# hFile, sWork
  PRINT# hFile, " UN-USED SUBS  (exported, or in incl. files, may be used by other programs)"
  PRINT# hFile, uline
  IF iFuncts THEN
     FOR I = 0 TO iFuncts - 1
        IF Functs(I).IsUsed = 1 AND Functs(I).iType = 1 THEN
            fName = Files(Functs(I).FileNum)
            aStr2 = ""
            IF Functs(I).Exported THEN aStr2 = " <EXPORT>"
            PRINT# hFile, LEFT$(" SUB " & Functs(i).zName & aStr2 & SPACE$(50), 50) & "    [" & _
                   MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" STR$(Functs(I).LineNum)
        END IF
     NEXT I
  END IF
'------------------------------------------------------------------------------
  IF DeclCount THEN
     PRINT# hFile, ""
     PRINT# hFile, sWork
     PRINT# hFile, " DECLARED, BUT NON-EXISTING SUB/FUNCTION(S)"
     PRINT# hFile, uline

     FOR I = 0 TO DeclCount - 1
        IF iFuncts > 0 THEN
           aStr = gDecl(I).uName + CHR$(0)
           ARRAY SCAN Functs(), FROM 1 TO LEN(aStr), = aStr, TO fDbl
        END IF
        IF fDbl = 0 THEN
           fName = Files(gDecl(I).FileNum)
           aStr2 = ""
           IF gDecl(I).iType = 1 THEN
              aStr = " SUB "
           ELSEIF gDecl(I).iType = 2 THEN
              aStr = " FUNCTION "
           ELSEIF gDecl(I).iType = 3 THEN
              aStr = " CALLBACK "
           END IF
           IF gDecl(I).Exported  THEN aStr2 = " <EXPORT>"
           PRINT# hFile, LEFT$(aStr & gDecl(I).zName & aStr2 & SPACE$(50), 50) & "    [" & _
                  MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" STR$(gDecl(I).LineNum)
        END IF
     NEXT I
  END IF
'------------------------------------------------------------------------------
  IF iFuncts THEN
     PRINT# hFile, ""
     PRINT# hFile, sWork
     PRINT# hFile, " EXISTING, BUT NON-DECLARED SUB/FUNCTION(S)"
     PRINT# hFile, uline

     FOR I = 0 TO iFuncts - 1
        IF DeclCount THEN
           aStr = Functs(I).uName + CHR$(0)
           ARRAY SCAN gDecl(), FROM 1 TO LEN(aStr), = aStr, TO fDbl
        END IF
        IF fDbl = 0 THEN
           SELECT CASE UCASE$(Functs(I).zName)
              CASE "PBMAIN", "WINMAIN", "LIBMAIN", "PBLIBMAIN", "DLLMAIN" 'ignore these
              CASE ELSE
                 fName = Files(Functs(I).FileNum)
                 aStr2 = ""
                 IF Functs(I).iType = 1 THEN
                    aStr = " SUB "
                 ELSEIF Functs(I).iType = 2 THEN
                    aStr = " FUNCTION "
                 ELSEIF Functs(I).iType = 3 THEN
                    aStr = " CALLBACK "
                 END IF
                 IF Functs(I).Exported  THEN aStr2 = " <EXPORT>"
                 PRINT# hFile, USING$("####", Functs(I).IsUsed - 1) & _
                        LEFT$(aStr & Functs(I).zName & aStr2 & SPACE$(45), 45) & "    [" & _
                        MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" STR$(Functs(I).LineNum)
              END SELECT
           END IF
        NEXT
  END IF
'------------------------------------------------------------------------------
  PRINT# hFile, ""
  PRINT# hFile, sWork
  PRINT# hFile, " UN-USED GLOBAL VARIABLES"
  PRINT# hFile, uline
  IF igVars THEN
     FOR I = 0 TO igVars - 1
        IF gVars(I).IsUsed = 1 THEN
           fName = Files(gVars(I).FileNum)
           PRINT# hFile, " " & LEFT$(gVars(i).zName & SPACE$(47), 47) & "    [" &  _
                  MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(gVars(I).LineNum)
        END IF
     NEXT I
  END IF
'------------------------------------------------------------------------------
  PRINT# hFile, ""
  PRINT# hFile, sWork
  PRINT# hFile, " UN-USED LOCAL VARIABLES"
  PRINT# hFile, uline
  IF ilVars THEN
     FOR I = 0 TO ilVars - 1
        IF lVars(I).IsUsed = 1 THEN
           fName = Files(lVars(I).FileNum)
           PRINT# hFile, " " & LEFT$(lVars(i).zName & SPACE$(47), 47) & "    [" &  _
                  MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(lVars(I).LineNum)
        END IF
     NEXT I
  END IF
'------------------------------------------------------------------------------
  IF igVars AND ilVars THEN
     FOR I = 0 TO igVars - 1
        aStr = gVars(I).uName & CHR$(0)
        ARRAY SCAN lVars(), FROM 1 TO LEN(aStr), = aStr, TO cTmp
        IF cTmp THEN EXIT FOR
     NEXT
     IF cTmp THEN
        PRINT# hFile, ""
        PRINT# hFile, sWork
        PRINT# hFile, " GLOBAL/LOCAL MIX - WARNING!"
        PRINT# hFile, " Following global variable name(s) exist in both global and local"
        PRINT# hFile, " form. While the compiler allows this, special care must be taken"
        PRINT# hFile, " to avoid hard-to-find errors. Please check them out carefully."
        PRINT# hFile, uline

        FOR I = 0 TO igVars - 1
           aStr = gVars(I).uName & CHR$(0)
           ARRAY SCAN lVars(), FROM 1 TO LEN(aStr), = aStr, TO lRes
           IF lRes THEN
              cTmp = 0 : fDbl = 0
              fName = Files(gVars(I).FileNum)
              PRINT# hFile, " " & LEFT$(gVars(I).zName & SPACE$(47), 47) & "    [" &  _
                            MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(gVars(I).LineNum)
              DO
                 cTmp = cTmp + lRes
                 fName = Files(lVars(cTmp - 1).FileNum)
                 PRINT# hFile, "   local in " &  _
                            MID$(fName, INSTR(-1, fName, "\") + 1) & " :" & STR$(lVars(cTmp - 1).LineNum)
                 ARRAY SCAN lVars(cTmp), FROM 1 TO LEN(aStr), = aStr, TO lRes
              LOOP WHILE lRes
           END IF
        NEXT
     END IF
  END IF
'------------------------------------------------------------------------------
  IF igDbl THEN
     PRINT# hFile, ""
     PRINT# hFile, sWork
     PRINT# hFile, " DUPLICATE GLOBAL NAMES - WARNING!"
     PRINT# hFile, " Following global name(s) exist as both array and varable."
     PRINT# hFile, " While the compiler allows this, special care must be taken"
     PRINT# hFile, " avoid hard-to-find errors. Please check them out carefully."
     PRINT# hFile, uline
     FOR I = 0 TO igDbl - 1
        fName = Files(gDbl(I).FileNum)
        PRINT# hFile, " " & LEFT$(gDbl(I).zName & SPACE$(47), 47) & "    [" &  _
                      MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(gDbl(I).LineNum)
     NEXT
  END IF
'------------------------------------------------------------------------------
  'code added by Wayne Diamond, slightly altered by Borje Hagsten
  PRINT# hFile, ""
  PRINT# hFile, sWork
  PRINT# hFile, " TOTAL REFERENCE COUNT - (Count, Name, [declared in File] : at Line number)"
  PRINT# hFile, " Lists how many times the following has been called/used (zero = un-used)"
  PRINT# hFile, uline
'------------------------------------------------------------------------------
  IF iFuncts > 0 THEN
     PRINT# hFile, " FUNCTIONS:"
     FOR I = 0 TO iFuncts - 1
        IF Functs(I).iType > 1 THEN
           SELECT CASE UCASE$(Functs(I).zName)
              CASE "PBMAIN", "WINMAIN", "LIBMAIN", "PBLIBMAIN", "DLLMAIN" 'ignore these
              CASE ELSE
                 fName = Files(Functs(I).FileNum)
                 aStr = ""
                 IF Functs(I).Exported THEN aStr = " <EXPORT>"
                 PRINT# hFile, USING$("####", Functs(I).IsUsed - 1) & "  " & _
                        LEFT$(Functs(I).zName & aStr & SPACE$(43), 43) & "    [" & _
                        MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(Functs(I).LineNum)
           END SELECT
        END IF
     NEXT I
  END IF
'------------------------------------------------------------------------------
  PRINT# hFile, ""
  IF iFuncts > 0 THEN
     PRINT# hFile, " SUBS:"
     FOR I = 0 TO iFuncts - 1
        IF Functs(I).iType = 1 THEN
           fName = Files(Functs(I).FileNum)
           aStr = ""
           IF Functs(I).Exported THEN aStr = " <EXPORT>"
           PRINT# hFile, USING$("####", Functs(I).IsUsed - 1) & "  " & _
                  LEFT$(Functs(I).zName & aStr & SPACE$(43), 43) & "    [" & _
                  MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(Functs(I).LineNum)
        END IF
      NEXT I
  END IF
'------------------------------------------------------------------------------
  PRINT# hFile, ""
  IF igVars > 0 THEN
     PRINT# hFile, " GLOBAL VARIABLES:"
     FOR I = 0 TO igVars - 1
        fName = Files(gVars(I).FileNum)
        PRINT# hFile, USING$("####", gVars(I).IsUsed - 1) & "  " & _
                      LEFT$(gVars(I).zName & SPACE$(43), 43) & "    [" & _
                      MID$(fName, INSTR(-1, fName, "\") + 1) & "] :" & STR$(gVars(I).LineNum)
     NEXT I
  END IF
  'end of Wayne Diamond code

'------------------------------------------------------------------------------
' STRING LITERALS
'------------------------------------------------------------------------------
  IF sStrCount THEN
     PRINT# hFile, ""
     PRINT# hFile, sWork
     PRINT# hFile, " STRING LITERALS"
     fName = ""

     FOR I = 0 TO sStrCount - 1
        aStr = Files(VAL(PARSE$(sString(I), $TAB, 1)))
        aStr = MID$(aStr, INSTR(-1, aStr, "\") + 1)
        IF aStr <> fName THEN
           fName = aStr
           PRINT# hFile, ""
           IF I THEN PRINT# hFile, uline
           PRINT# hFile, " Line  Text     [" + fName + "]"
           PRINT# hFile, uline
        END IF
        PRINT# hFile,  PARSE$(sString(I), $TAB, 2) + "  " + _
                       PARSE$(sString(I), $TAB, 3)
     NEXT
  END IF
'------------------------------------------------------------------------------

  CLOSE hFile

  SLEEP 20
  ShellExecute 0, "open", FilePathStr + DestFile, BYVAL 0, BYVAL 0, %SW_SHOWNORMAL
  'launch log file
END SUB

'************************************************************************
' fExist - make sure a file or folder exists
'************************************************************************
FUNCTION fExist(BYVAL sfName AS STRING) AS LONG
  LOCAL hRes AS DWORD, tWFD AS WIN32_FIND_DATA

  IF ASC(sfName, -1) = 92 THEN                'if a path with trailing backslash
     sfName = LEFT$(sfName, LEN(sfName) - 1)  'remove trailing backslash
  END IF

  hRes = FindFirstFile(BYVAL STRPTR(sfName), tWFD)
  IF hRes <> %INVALID_HANDLE_VALUE THEN FUNCTION = %TRUE
  FindClose hRes
END FUNCTION

'************************************************************************
' GetCommandFile - loads a received Path&File name into global array
'************************************************************************
FUNCTION GetCommandFile(BYVAL CmdStr AS STRING, Fi() AS STRING) AS LONG
    LOCAL tmpName AS STRING, pStr AS STRING

     CmdStr = TRIM$(CmdStr)                   'trim away ev. leading/ending spaces
     IF LEFT$(CmdStr, 1) = CHR$(34) THEN      'if in double-quotes
       CmdStr = MID$(CmdStr, 2)               'remove first quote
       pStr = CHR$(34)                        'and use DQ as delimiter for PARSE$
    ELSE
       pStr = " "                             'else use space as delimiter
    END IF

    tmpName = TRIM$(PARSE$(CmdStr, pStr, 1))
    IF LEN(tmpName) = 0 THEN EXIT FUNCTION

    IF (GETATTR(tmpName) AND 16) = 0 THEN     'make sure it isn't a folder
       Fi(0) = tmpName
    ELSE
       EXIT FUNCTION
    END IF

    FUNCTION = 1                             'return number of collected files
END FUNCTION

'************************************************************************
' GetDroppedFile - Function Loads File/Folder names into the global arrays
'************************************************************************
FUNCTION GetDroppedFile( BYVAL hfInfo AS LONG, Fi() AS STRING) AS LONG
  LOCAL Count AS LONG, ln AS LONG, tmp AS STRING, fName AS ASCIIZ * %MAX_PATH

  Count = DragQueryFile(hfInfo, &HFFFFFFFF&, BYVAL %NULL, 0) 'get number of dropped files

  IF Count THEN                                          'If we got something
     ln = DragQueryFile(hfInfo, 0, fName, %MAX_PATH)     'put FileName into fString And get len
     IF ln THEN
        tmp = TRIM$(LEFT$(fName, ln))
        IF LEN(tmp) AND (GETATTR(tmp) AND 16) = 0 THEN   'make sure it's a file, not a folder
           Fi(0) = tmp
           FUNCTION = 1
        END IF
     END IF
  END IF

  CALL DragFinish(hfInfo)
END FUNCTION

'************************************************************************
' Get PB/DLL 6 compiler's include dir (winapi folder)
'************************************************************************
FUNCTION GetIncludeDir AS STRING
  LOCAL lRet   AS LONG, hKey AS LONG
  LOCAL Buffer AS ASCIIZ * %MAX_PATH, SubKey AS STRING

  Buffer = "Software\PowerBASIC\PB/Win\7.00"
  SubKey = "Filename"
  IF RegOpenKeyEx(%HKEY_LOCAL_MACHINE, Buffer, 0, _
                  %KEY_QUERY_VALUE, hKey) = %ERROR_SUCCESS THEN

     lRet = RegQueryValueEx(hKey, BYVAL STRPTR(SubKey), _
            BYVAL 0&, BYVAL 0&, Buffer, SIZEOF(Buffer))

     IF LEN(Buffer) THEN FUNCTION = TRIM$(Buffer)
     IF hKey THEN RegCloseKey hKey

     IF LEN(TRIM$(Buffer)) THEN
        Buffer = TRIM$(Buffer)
        Buffer = LEFT$(Buffer, INSTR(-1, Buffer, ANY "\/"))           ' Compiler path
        SubKey = LEFT$(Buffer, INSTR(-1, Buffer, "\Bin\")) + "WinAPI" ' WinAPI path
        Buffer = IniGetString("Compiler", "Include0", SubKey, Buffer + "PBWin.ini")
        IF LEN(TRIM$(Buffer)) THEN
           Buffer = TRIM$(Buffer)
           IF RIGHT$(Buffer, 1) <> "\" THEN
              IF RIGHT$(Buffer, 1) <> "/" THEN Buffer = Buffer + "\"
           END IF
           FUNCTION = TRIM$(Buffer)
           EXIT FUNCTION
        END IF
     END IF
  END IF

  Buffer = "Software\PowerBasic\PB/DLL\6.10\Compiler"
  SubKey = "Include"
  IF RegOpenKeyEx(%HKEY_CURRENT_USER, Buffer, 0, _
                  %KEY_QUERY_VALUE, hKey) = %ERROR_SUCCESS THEN

     lRet = RegQueryValueEx(hKey, BYVAL STRPTR(SubKey), _
            BYVAL 0&, BYVAL 0&, Buffer, SIZEOF(Buffer))

     IF LEN(Buffer) THEN FUNCTION = TRIM$(Buffer)
'     MSGBOX Buffer
     FUNCTION = Buffer
     IF hKey THEN RegCloseKey hKey

  ELSE
     Buffer = "Software\PowerBasic\PB/DLL\6.00\Compiler"
     IF RegOpenKeyEx(%HKEY_CURRENT_USER, Buffer, 0, _
                     %KEY_QUERY_VALUE, hKey) = %ERROR_SUCCESS THEN

        lRet = RegQueryValueEx(hKey, BYVAL STRPTR(SubKey), _
               BYVAL 0&, BYVAL 0&, Buffer, SIZEOF(Buffer))

        IF LEN(Buffer) THEN FUNCTION = TRIM$(Buffer)
        IF hKey THEN RegCloseKey hKey
     END IF
  END IF
END FUNCTION

'
' Get string from ini file
'
FUNCTION IniGetString(BYVAL sSection AS STRING, BYVAL sKey AS STRING, _
                      BYVAL sDefault AS STRING, BYVAL sFile AS STRING) AS STRING
  LOCAL RetVal AS LONG, zResult AS ASCIIZ * %MAX_PATH

  RetVal = GetPrivateProfileString(BYVAL STRPTR(sSection), _
                                   BYVAL STRPTR(sKey), _
                                   BYVAL STRPTR(sDefault), _
                                   zResult, SIZEOF(zResult), BYVAL STRPTR(sFile))
  IF RetVal THEN FUNCTION = TRIM$(LEFT$(zResult, RetVal))
END FUNCTION

'
' Check to see if the file has a #COMPILE metastatement
'
FUNCTION IsFileMain(BYVAL fName AS STRING) AS LONG
  LOCAL hFile AS LONG, txt AS STRING

  hFile = FREEFILE                           'get a free file handle
  OPEN fName FOR INPUT AS hFile LEN = 16383  'open file

  IF ERR THEN                                'if it failed
     RESET : ERRCLEAR                        'reset, clear error
     FUNCTION = -2 : EXIT FUNCTION           'return -2 to indicate failure and exit
  END IF

  IF LOF(hFile) = 0 THEN                     'if zero length file
     CLOSE hFile                             'close it
     FUNCTION = -3 : EXIT FUNCTION           'return -3 to indicate empty file and exit
  END IF

  DO WHILE EOF(hFile) = 0                    'loop through file
     LINE INPUT# hFile, txt                  'line by line
     IF LEN(txt) > 8 THEN                    'if enough long
        txt = UCASE$(LTRIM$(txt))            'adjust txt for compare
        IF (ASC(txt) = 35 OR ASC(txt) = 36) AND MID$(txt, 2, 8) = "COMPILE " THEN
           FUNCTION = -1
           EXIT DO
        END IF

        IF LEFT$(txt, 9) = "FUNCTION " OR _  'jump out once we hit a Sub or Function
                LEFT$(txt, 4) = "SUB " OR _
                   LEFT$(txt, 9) = "CALLBACK " OR _
                      LEFT$(txt, 7) = "STATIC " THEN

           FUNCTION = LOF(hFile) 'return length
           EXIT DO
        END IF
     END IF
  LOOP
  CLOSE hFile

END FUNCTION

'
' UCASER function, returns UCASE string without altering original
'
FUNCTION UCASER(BYVAL st AS STRING) AS STRING
  #REGISTER NONE
  LOCAL p AS STRING PTR
  p = STRPTR(st)

  ! mov eax, p              ; move pointer to string into eax
  ! mov ecx, [eax-4]        ; move length of string into ecx (counter)
  ! cmp ecx, 0              ; if length is 0, no string length
  ! je exitUCASER           ; then exit

  beginUCASER:
     ! mov dl, [eax]        ; move current char into dl
     ! cmp dl, 97           ; compare against value 97 (a)
     ! jb nextUCASER        ; if dl < 97  then get next character
     ! cmp dl, 123          ; compare against value 123
     ! jb makeUCASER        ; if dl < 123 it is in 97-122 range, make Uppercase and get next
     ! cmp dl, 224          ; compare against value 224 () - extended ANSI
     ! jb nextUCASER        ; if dl < 224 it is in 123-224 range, do nothing to it
     ! cmp dl, 247          ; compare against value 247
     ! jb makeUCASER        ; if dl < 247 it is in 224-247 range, make Uppercase and get next
     ! je nextUCASER        ; if dl = 247, do nothing
     ! cmp dl, 255          ; compare against value 255
     ! jb makeUCASER        ; if dl < 255 it is in 248-255 range, make Uppercase and get next
     ! jmp nextUCASER       ; else, on to next character

  makeUCASER:
     ! sub dl, 32           ; make lowercase by adding 32 to dl's value
     ! mov [eax], dl        ; write changed char back into eax and fall through to nextUCASER

  nextUCASER:
     ! inc eax               ; get next character
     ! dec ecx               ; decrease ecx (length) counter
     ! jnz beginUCASER       ; iterate if not zero (end of string)
     FUNCTION = st

  exitUCASER:
END FUNCTION

'
' UCASES routine, alters given string directly - very fast..
'
SUB UCASES(a$)
  IF LEN(a$) = 0 THEN EXIT SUB
  #REGISTER NONE

  ! mov eax, a$            ; eax = pointer to string handle
  ! mov eax, [eax]         ; eax = pointer to string data
  ! mov ecx, [eax-4]       ; move length of string into ecx

  BgnUpLoop:
     ! mov dl, [eax]       ; move current char into dl
     ! cmp dl, 97          ; compare against value 97 (a)
     ! jb NoDnChar         ; if dl < 97 it is in 0-97 range, do nothing - get next
     ! cmp dl, 123         ; compare against value 123
     ! jb MakeUpper        ; if dl < 123 it is in 97-122 range, make Upper case and get next
     ! cmp dl, 224         ; compare against value 224 () - extended ANSI characters
     ! jb NoDnChar         ; if dl < 224 it is in 123-223 range, do nothing - get next
     ! cmp dl, 247         ; compare against value 247 ()
     ! jb MakeUpper        ; if dl < 247 it is in 224-246 range, make Upper case and get next
     ! je NoDnChar         ; else dl = 247 (), do nothing - get next
     ! cmp dl, 255         ; compare against value 255 ()
     ! jb MakeUpper        ; if dl < 255 it is in 248-254 range, make Upper case and get next
     ! jmp NoDnChar        ; else dl = 255, do nothing - get next

  MakeUpper:
     ! sub dl, 32          ; make uppercase subtracting 32 from dl's value
     ! mov [eax], dl       ; write changed char back into a$

  NoDnChar:
     ! inc eax             ; get next character
     ! dec ecx             ; decrease ecx (length) counter
     ! jnz bgnUpLoop       ; iterate if not zero
END SUB
